import functools
import hashlib

signing_algs = (
    ('RSASSA_PSS_SHA_256',        'PS256', 'sha256'),
    ('RSASSA_PSS_SHA_384',        'PS384', 'sha384'),
    ('RSASSA_PSS_SHA_512',        'PS512', 'sha512'),
    ('RSASSA_PKCS1_V1_5_SHA_256', 'RS256', 'sha256'),
    ('RSASSA_PKCS1_V1_5_SHA_384', 'RS384', 'sha384'),
    ('RSASSA_PKCS1_V1_5_SHA_512', 'RS512', 'sha512'),
    ('ECDSA_SHA_256',             'ES256', 'sha256'),
    ('ECDSA_SHA_384',             'ES384', 'sha384'),
    ('ECDSA_SHA_512',             'ES512', 'sha512'),
    )

jwa2aws = { jwa: aws for aws, jwa, _ in signing_algs }
aws2jwa = { aws: jwa for aws, jwa, _ in signing_algs }
jwa2halg = { jwa: functools.partial(hashlib.new, halg) for _, jwa, halg in signing_algs }



import base64
import json

from . import jwa, jwk


class JWS:
    def __init__(self, payload=None):
        self.payload = json.dumps(payload).encode('utf-8')
        self.signatures = list()

    def add_signature(self, key, alg='RS256', protected=None, header=None):
        if key.use != 'sig':
            raise jwk.JWKError('The key is not for signing')

        if header is not None:
            raise NotImplementedError('Unprotected header not implemented')

        try:
            aws_alg = jwa.jwa2aws[alg]
        except KeyError:
            raise jwk.JWKError('Algorithm {} not possible'.format(alg))

        if alg in ('ES256', 'ES385', 'ES512'):
            raise NotImplementedError('DER to Jose signature conversion not yet implemented')

        if protected is None:
            protected = dict()

        if header is None:
            header = dict()

        protected_header = dict(
            typ='JWS',
            alg=alg,
            kid=key.public_key_jwk['kid'],
            )
        protected_header.update(protected)
        protected_header = json.dumps(protected_header).encode('utf-8')

        signing_input = jwa.jwa2halg[alg](
            base64.urlsafe_b64encode(protected_header).rstrip(b'=')
            + b'.'
            + base64.urlsafe_b64encode(self.payload).rstrip(b'=')
        ).digest()

        signature = key.sign_digest(signing_input, aws_alg)

        self.signatures.append(
            dict(
                protected=base64.urlsafe_b64encode(protected_header).decode('utf-8').rstrip('='),
                signature=base64.urlsafe_b64encode(signature).decode('utf-8').rstrip('=')
                ))

        return self

    def serialize(self, compact=False):
        payload = base64.urlsafe_b64encode(self.payload).decode('utf-8').rstrip('=')

        if compact:
            if len(self.signatures) > 1:
                raise jwk.JWKError('Too many signatures for compact JWT')
            if not self.signatures:
                raise jwk.JWKError('Not signed, can\'t serialize JWT')
            return '{}.{}.{}'.format(
                self.signatures[0]['protected'],
                payload,
                self.signatures[0]['signature']
                )

        token = dict(
            payload=payload,
            signatures=self.signatures
            )

        return json.dumps(token)

    @classmethod
    def from_jose_token(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jwt_kms']

package_data = \
{'': ['*']}

install_requires = \
['cryptography>=37.0.4,<38.0.0']

entry_points = \
{'console_scripts': ['test = scripts:test']}

setup_kwargs = {
    'name': 'jwt-kms',
    'version': '0.1.0',
    'description': 'Library to offload some JWT crypto operations to KMS',
    'long_description': None,
    'author': 'Juha-Matti Tapio',
    'author_email': 'jmtapio@verkkotelakka.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import base64
import cryptography.hazmat.primitives.serialization
import functools
import hashlib
import json
import math

from . import jwa


class JWKError(Exception):
    pass


class Key:
    def __init__(self, client, keyId):
        self.client = client
        self.keyId = keyId

    @functools.cached_property
    def key_info(self):
        return self.client.get_public_key(KeyId=self.keyId)

    @property
    def public_key(self):
        return cryptography.hazmat.primitives.serialization.load_der_public_key(
            self.key_info['PublicKey'])

    @property
    def public_key_pem(self):
        return self.public_key.public_bytes(
            cryptography.hazmat.primitives.serialization.Encoding.PEM,
            cryptography.hazmat.primitives.serialization.PublicFormat.SubjectPublicKeyInfo
        )

    @property
    def public_key_jwk(self):
        if self.keyspec.startswith('RSA_'):
            num = self.public_key.public_numbers()
            core = dict(
                kty='RSA',
                n=self._encode_b64u(num.n),
                e=self._encode_b64u(num.e)
                )
            core.update(dict(
                use=self.use,
                kid=self._thumbprint(core),
                ))
            return core

        elif self.keyspec.startswith('ECC_NIST_'):
            num = self.public_key.public_numbers()
            core = dict(
                kty='EC',
                crv={
                    'ECC_NIST_P256': 'P-256',
                    'ECC_NIST_P384': 'P-384',
                    'ECC_NIST_P512': 'P-512',
                }[self.keyspec],
                x=self._encode_b64u(num.x),
                y=self._encode_b64u(num.y),
            )
            core.update(dict(
                use=self.use,
                kid=self._thumbprint(core),
                ))
            return core

        raise JWKError('Unknown keyspec {}'.format(self.keyspec))

    @property
    def use(self):
        return {
            'ENCRYPT_DECRYPT': 'enc',
            'SIGN_VERIFY': 'sig',
        }.get(self.key_info.get('KeyUsage'))

    @property
    def supported_algs(self):
        return self.key_info.get('SigningAlgorithms', [])

    @property
    def keyspec(self):
        return self.key_info.get('CustomerMasterKeySpec', '')

    def sign_digest(self, digest, alg):
        if alg not in self.supported_algs:
            raise JWKError('Alg {} not supported by this key. Supported: {}'.format(
                alg, self.supported_algs))

        return self.client.sign(
            KeyId=self.keyId,
            Message=digest,
            MessageType='DIGEST',
            SigningAlgorithm=alg
        )['Signature']

    def _encode_b64u(self, number):
        return base64.urlsafe_b64encode(
            number.to_bytes(
                math.ceil(math.log2(number)/8), 'big'
                )
        ).decode('utf-8').rstrip('=')

    def _thumbprint(self, d):
        return base64.urlsafe_b64encode(
            hashlib.sha256(
                json.dumps(d, sort_keys=True, separators=(',', ':')).encode('utf-8')
            ).digest()
        ).decode('utf-8').rstrip('=')
